from nbodykit.base.catalog import CatalogSourceBase, column
from nbodykit.utils import attrs_to_dict

import numpy
import logging
import functools

def OnDemandColumn(source, col):
    """
    Return a column from the source on-demand.
    """
    return source[col]

class MultipleSpeciesCatalog(CatalogSourceBase):
    """
    A CatalogSource interface for handling multiples species
    of particles.

    Parameters
    ----------
    names : list of str
        list of strings specifying the names of the various species;
        data columns are prefixed with "species/" where "species" is
        in ``names``
    *species : two or more CatalogSource objects
        catalogs to be combined into a single catalog, which give the
        data for different species of particles; as many catalogs
        as names must be provided
    use_cache : bool, optional
        whether to cache data when reading; default is ``True``

    Examples
    --------
    >>> source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42)
    >>> source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84)
    >>> cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)
    """
    logger = logging.getLogger('MultipleSpeciesCatalog')

    def __repr__(self):
        return "MultipleSpeciesCatalog(species=%s)" %str(self.attrs['species'])

    def __init__(self, names, *species, **kwargs):

        # whether to use the cache
        use_cache = kwargs.pop('use_cache', True)

        # input checks
        if len(species) < 2:
            raise ValueError("need at least 2 particle species to initialize MultipleSpeciesCatalog")
        if len(set(names)) != len(names):
            raise ValueError("each species must have a unique name")
        if not all(cat.comm is species[0].comm for cat in species):
            raise ValueError("communicator mismatch in MultipleSpeciesCatalog")
        if len(names) != len(species):
            raise ValueError("a name must be provided for each species catalog provided")

        self.comm = species[0].comm
        self.species = names
        self.attrs['species'] = names

        # update the dictionary with data/randoms attrs
        for cat, name in zip(species, names):
            self.attrs.update(attrs_to_dict(cat, name + '.'))

        # update the rest of meta-data
        self.attrs.update(kwargs)

        # no size!
        self.size = NotImplemented

        # store the local sizes of each species catalog
        self._sizes = [source.size for source in species]

        # init the base class
        CatalogSourceBase.__init__(self, self.comm, use_cache=use_cache)

        # turn on cache?
        if self.use_cache:
            for cat in species: cat.use_cache = True

        # prefixed columns in this source return on-demand from their
        # respective source objects
        for name, source in zip(names, species):
            for col in source.columns:
                f = functools.partial(OnDemandColumn, col=col, source=source)
                self._overrides[name+'/'+col] = f

    def __getitem__(self, key):
        """
        This modifies the behavior of :func:`CatalogSourceBase.__getitem__`
        such that if ``key`` is a species name, a
        :class:`~nbodykit.base.catalog.CatalogCopy` will be returned that
        holds that data only for the species.
        """
        # return a new CatalogSource holding only the specific species
        if key in self.species:
            from nbodykit.base.catalog import CatalogCopy

            # get the data columns for this species
            data = {}
            for col in self:
                if col.startswith(key):
                    name = col.split('/')[-1]
                    data[name] = self[col]

            # size of the underlying source
            size = self._sizes[self.species.index(key)]

            # the returned object
            toret = CatalogCopy(size, self.comm, use_cache=self.use_cache, **data)

            # copy over the meta data
            for k in self.attrs:
                if k.startswith(key+'.'):
                    toret.attrs[k[len(key)+1:]] = self.attrs[k]

            return toret

        # base class __getitem__
        return CatalogSourceBase.__getitem__(self, key)

    def __setitem__(self, col, value):
        """
        Add columns to any of the species catalogs.

        .. note::
            New column names should be prefixed by 'species/' where
            'species' is a name in the :attr:`species` attribute.
        """
        fields = col.split('/')
        if len(fields) != 2:
            msg = "new column names should be prefixed by 'species/' where "
            msg += "'species' is one of %s" % str(self.species)
            raise ValueError(msg)

        if fields[0] not in self.species:
            args = (fields[0], str(self.species))
            raise ValueError("species '%s' is not valid; should be one of %s" %args)

        # check size
        size = self._sizes[self.species.index(fields[0])]
        if not numpy.isscalar(value):
            assert len(value) == size, "error setting '%s' column, data must be array of size %d" % (col,size)

        return CatalogSourceBase.__setitem__(self, col, value)

    def to_mesh(self, Nmesh=None, BoxSize=None, dtype='f4', interlaced=False,
                compensated=False, window='cic', weight='Weight',
                selection='Selection', value='Value', position='Position'):
        """
        Convert the catalog to a mesh, which knows how to "paint" the
        the combined density field, summed over all particle species.

        Parameters
        ----------
        Nmesh : int, 3-vector, optional
            the number of cells per box side; can be inferred from ``attrs``
            if the value is the same for all species
        BoxSize : float, 3-vector, optional
            the size of the box; can be inferred from ``attrs``
            if the value is the same for all species
        dtype : str, dtype, optional
            the data type of the mesh when painting
        interlaced : bool, optional
            whether to use interlacing to reduce aliasing when painting the
            particles on the mesh
        compensated : bool, optional
            whether to apply a Fourier-space transfer function to account for
            the effects of the gridding + aliasing
        window : str, optional
            the string name of the window to use when interpolating the
        weight : str, optional
            the name of the column specifying the weight for each particle
        value: str, optional
            the name of the column specifying the field value for each particle
        selection : str, optional
            the name of the column that specifies which (if any) slice
            of the CatalogSource to take
        position : str, optional
            the name of the column that specifies the position data of the
            objects in the catalog
        """
        from nbodykit.source.catalogmesh import MultipleSpeciesCatalogMesh

        # verify that all of the required columns exist
        for name in self.species:
            for col in [position, selection, weight, value]:
                _col = name+'/'+col
                if _col not in self:
                    raise ValueError("the '%s' species is missing the '%s' column" %(name, col))

        # try to find BoxSize and Nmesh
        if BoxSize is None:
            BoxSize = check_species_metadata('BoxSize', self.attrs, self.species)
        if Nmesh is None:
            Nmesh = check_species_metadata('Nmesh', self.attrs, self.species)


        # initialize the mesh
        kws = {'Nmesh':Nmesh, 'BoxSize':BoxSize, 'dtype':dtype, 'selection':selection}
        mesh = MultipleSpeciesCatalogMesh(self,
                                          Nmesh=Nmesh,
                                          BoxSize=BoxSize,
                                          dtype=dtype,
                                          selection=selection,
                                          position=position,
                                          weight=weight,
                                          value=value)
        mesh.interlaced = interlaced
        mesh.compensated = compensated
        mesh.window = window

        return mesh

def check_species_metadata(name, attrs, species):
    """
    Check to see if there is a single value for
    ``name`` in the meta-data of all the species
    """
    _vals = []

    if name in attrs:
        _vals.append(attrs[name])

    for s in species:
        if s + '.' + name in attrs:
            _vals.append(attrs.get(s + '.' + name ))

    if len(_vals) == 0:
        raise ValueError("please specify ``%s`` attributes")

    if not all(numpy.equal(_vals[0], i).all() for i in _vals):
        raise ValueError("please specify ``%s`` attributes that are consistent for each species and for the multi species catalog; " % name)

    return _vals[0]



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from nbodykit.transform import ConstantArray
import pytest

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_sky_to_cartesian(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # make source
    s = RandomCatalog(csize=100, seed=42)

    # ra, dec, z
    s['z']   = s.rng.normal(loc=0.5, scale=0.1, size=s.size)
    s['ra']  = s.rng.uniform(low=110, high=260, size=s.size)
    s['dec'] = s.rng.uniform(low=-3.6, high=60., size=s.size)

    # make the position array
    s['Position1'] = transform.SkyToCartesion(s['ra'], s['dec'], s['z'], cosmo)

    # requires dask array
    with pytest.raises(TypeError):
        s['Position1'] = transform.SkyToCartesion(s['ra'].compute(), s['dec'], s['z'], cosmo)

@MPITest([1, 4])
def test_stack_columns(comm):
    CurrentMPIComm.set(comm)

    # make source
    s = RandomCatalog(csize=100, seed=42)

    # add x,y,z
    s['x'] = s.rng.uniform(0, 2600., size=s.size)
    s['y'] = s.rng.uniform(0, 2600., size=s.size)
    s['z'] = s.rng.uniform(0, 2600., size=s.size)

    # stack
    s['Position'] = transform.StackColumns(s['x'], s['y'], s['z'])

    # test equality
    x, y, z = s.compute(s['x'], s['y'], s['z'])
    pos = numpy.vstack([x,y,z]).T
    numpy.testing.assert_array_equal(pos, s['Position'])

    # requires dask array
    with pytest.raises(TypeError):
        s['Position'] = transform.StackColumns(x,y,z)

@MPITest([1, 4])
def test_combine(comm):
    CurrentMPIComm.set(comm)

    # make two sources
    s1 = UniformCatalog(3e-6, 2600)
    s2 = UniformCatalog(3e-6, 2600)

    # concatenate all columns
    cat = transform.ConcatenateSources(s1, s2)

    # check the size and columns
    assert cat.size == s1.size + s2.size
    assert set(cat.columns) == set(s1.columns)

    # only one column
    cat = transform.ConcatenateSources(s1, s2, columns='Position')
    pos = numpy.concatenate([s1['Position'], s2['Position']], axis=0)
    numpy.testing.assert_array_equal(pos, cat['Position'])

    # fail on invalid column
    with pytest.raises(ValueError):
        cat = transform.ConcatenateSources(s1, s2, columns='InvalidColumn')

@MPITest([1])
def test_constarray(comm):
    a = ConstantArray(1.0, 1, chunks=1000)
    assert len(a) == 1
    print(a.shape)
    assert a.shape == (1,)
    a = ConstantArray([1.0, 1.0], 1, chunks=1000)
    assert a.shape == (1, 2)

    a = ConstantArray([1.0, 1.0], 3, chunks=1000)
    assert a.shape == (3, 2)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import pytest

# debug logging
setup_logging("debug")

@MPITest([4])
def test_tsc_interlacing(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # interlacing with TSC
    mesh = source.to_mesh(window='tsc', Nmesh=64, interlaced=True, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)

@MPITest([4])
def test_cic_interlacing(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # interlacing with TSC
    mesh = source.to_mesh(window='cic', Nmesh=64, interlaced=True, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)



from nbodykit.base.mesh import MeshSource
from nbodykit.base.catalog import CatalogSource
from six import add_metaclass
import abc
import numpy
import logging

# for converting from particle to mesh
from pmesh import window
from pmesh.pm import RealField, ComplexField

class CatalogMesh(MeshSource, CatalogSource):
    """
    A class to convert a CatalogSource to a MeshSource, by interpolating
    the position of the discrete particles on to a mesh.

    Parameters
    ----------
    source : CatalogSource
        the input catalog that we wish to interpolate to a mesh
    BoxSize :
        the size of the box
    Nmesh : int, 3-vector
        the number of cells per mesh side
    dtype : str
        the data type of the values stored on mesh
    weight : str
        column in ``source`` that specifies the weight value for each
        particle in the ``source`` to use when gridding
    value : str
        column in ``source`` that specifies the field value for each particle;
        the mesh stores a weighted average of this column
    selection : str
        column in ``source`` that selects the subset of particles to grid
        to the mesh
    position : str, optional
        column in ``source`` specifying the position coordinates; default
        is ``Position``
    """
    logger = logging.getLogger('CatalogMesh')
    def __repr__(self):
        return "(%s as CatalogMesh)" % repr(self.source)

    # intended to be used by CatalogSource internally
    def __init__(self, source, BoxSize, Nmesh, dtype, weight,
                    value, selection, position='Position'):

        # ensure self.comm is set, though usually already set by the child.
        self.comm = source.comm
        self.source = source
        self.position = position
        self.selection = selection
        self.weight = weight
        self.value = value

        self.attrs.update(source.attrs)

        # this will override BoxSize and Nmesh carried from the source, if there is any!
        MeshSource.__init__(self, BoxSize=BoxSize, Nmesh=Nmesh, dtype=dtype, comm=source.comm)
        CatalogSource.__init__(self, comm=source.comm)

        # copy over the overrides
        self._overrides.update(self.source._overrides)

        self.attrs['position'] = self.position
        self.attrs['selection'] = self.selection
        self.attrs['weight'] = self.weight
        self.attrs['value'] = self.value
        self.attrs['compensated'] = True
        self.attrs['interlaced'] = False
        self.attrs['window'] = 'cic'

    @property
    def size(self):
        """
        The number of local particles.
        """
        return self.source.size

    @property
    def hardcolumns (self):
        """
        The names of the hard-coded columns in the source.
        """
        return self.source.hardcolumns

    def get_hardcolumn(self, col):
        """
        Return a hard-coded column
        """
        return self.source.get_hardcolumn(col)

    @property
    def interlaced(self):
        """
        Whether to use interlacing when interpolating the density field.

        See :ref:`the documentation <interlacing>` for further details.

        See also: Section 3.1 of `Sefusatti et al. 2015 <https://arxiv.org/abs/1512.07295>`_
        """
        return self.attrs['interlaced']

    @interlaced.setter
    def interlaced(self, interlaced):
        self.attrs['interlaced'] = interlaced

    @property
    def window(self):
        """
        String specifying the name of the interpolation kernel when
        gridding the density field.

        See :ref:`the documentation <window-kernel>` for further details.

        .. note::
            Valid values must be in :attr:`pmesh.window.methods`
        """
        return self.attrs['window']

    @window.setter
    def window(self, value):
        assert value in window.methods
        self.attrs['window'] = value

    @property
    def compensated(self):
        """
        Boolean flag to indicate whether to correct for the windowing
        kernel introduced when interpolating the discrete particles to
        a continuous field.

        See :ref:`the documentation <compensation>` for further details.
        """
        return self.attrs['compensated']

    @compensated.setter
    def compensated(self, value):
        self.attrs['compensated'] = value

    def to_real_field(self, out=None, normalize=True):
        """
        Paint the density field, by interpolating the position column
        on to the mesh.

        This computes the following meta-data attributes in the process of
        painting, returned in the :attr:`attrs` attributes of the returned
        RealField object:

        - N : int
            the (unweighted) total number of objects painted to the mesh
        - W : float
            the weighted number of total objects, equal to the collective
            sum of the 'weight' column
        - shotnoise : float
            the Poisson shot noise, equal to the volume divided by ``N``
        - num_per_cell : float
            the mean number of weighted objects per cell

        .. note::

            The density field on the mesh is normalized as :math:`1+\delta`,
            such that the collective mean of the field is unity.

        See the :ref:`documentation <painting-mesh>` on painting for more 
        details on painting catalogs to a mesh.

        Returns
        -------
        real : :class:`pmesh.pm.RealField`
            the painted real field; this has a ``attrs`` dict storing meta-data
        """
        # check for 'Position' column
        if self.position not in self:
            msg = "in order to paint a CatalogSource to a RealField, add a "
            msg += "column named '%s', representing the particle positions" %self.position
            raise ValueError(msg)

        pm = self.pm

        fullsize = 0 # track how many were selected out
        Nlocal = 0 # (unweighted) number of particles read on local rank
        Wlocal = 0 # (weighted) number of particles read on local rank

        # the paint brush window
        paintbrush = window.methods[self.window]

        # initialize the RealField to returns
        if out is not None:
            assert isinstance(out, RealField), "output of to_real_field must be a RealField"
            numpy.testing.assert_array_equal(out.pm.Nmesh, pm.Nmesh)
            real = out
        else:
            real = RealField(pm)
            real[:] = 0

        # need 2nd field if interlacing
        if self.interlaced:
            real2 = RealField(pm)
            real2[...] = 0

        # read the necessary data (as dask arrays)
        columns = [self.position, self.weight, self.value, self.selection]
        Position, Weight, Value, Selection = self.read(columns)

        # ensure the slices are synced, since decomposition is collective
        N = max(pm.comm.allgather(len(Position)))

        # paint data in chunks on each rank
        chunksize = 1024 ** 2
        for i in range(0, N, chunksize):
            s = slice(i, i + chunksize)

            if len(Position) != 0:

                # be sure to use the source to compute
                position, weight, value, selection = \
                    self.source.compute(Position[s], Weight[s], Value[s], Selection[s])
            else:
                # workaround a potential dask issue on empty dask arrays
                position = numpy.empty((0, 3), dtype=Position.dtype)
                weight = None
                value = None
                selection = None

            if weight is None:
                weight = numpy.ones(len(position))

            if value is None:
                value = numpy.ones(len(position))

            # track all particles, before Selection applied
            fullsize += len(position)

            # apply any Selections
            if selection is not None:
                position = position[selection]
                weight = weight[selection]
                value = value[selection]

            # track total (selected) number and sum of weights
            Nlocal += len(position)
            Wlocal += weight.sum()

            # no interlacing
            if not self.interlaced:
                lay = pm.decompose(position, smoothing=0.5 * paintbrush.support)
                p = lay.exchange(position)
                w = lay.exchange(weight)
                v = lay.exchange(value)
                real.paint(p, mass=w * v, resampler=paintbrush, hold=True)

            # interlacing: use 2 meshes separated by 1/2 cell size
            else:
                lay = pm.decompose(position, smoothing=1.0 * paintbrush.support)
                p = lay.exchange(position)
                w = lay.exchange(weight)
                v = lay.exchange(value)

                H = pm.BoxSize / pm.Nmesh

                # in mesh units
                shifted = pm.affine.shift(0.5)

                # paint to two shifted meshes
                real.paint(p, mass=w * v, resampler=paintbrush, hold=True)
                real2.paint(p, mass=w * v, resampler=paintbrush, transform=shifted, hold=True)
                c1 = real.r2c()
                c2 = real2.r2c()

                # and then combine
                for k, s1, s2 in zip(c1.slabs.x, c1.slabs, c2.slabs):
                    kH = sum(k[i] * H[i] for i in range(3))
                    s1[...] = s1[...] * 0.5 + s2[...] * 0.5 * numpy.exp(0.5 * 1j * kH)

                c1.c2r(real)

        # unweighted number of objects
        N = pm.comm.allreduce(Nlocal)

        # weighted number of objects
        W = pm.comm.allreduce(Wlocal)

        # the full size; should be equal to csize
        fullsize = pm.comm.allreduce(fullsize)

        # weighted number density (objs/cell)
        nbar = 1. * W / numpy.prod(pm.Nmesh)

        # make sure we painted something!
        if N == 0:
            raise ValueError(("trying to paint particle source to mesh, "
                              "but no particles were found!"))

        # shot noise is volume / un-weighted number
        shotnoise = numpy.prod(pm.BoxSize) / N

        # save some meta-data
        real.attrs = {}
        real.attrs['shotnoise'] = shotnoise
        real.attrs['N'] = N
        real.attrs['W'] = W
        real.attrs['num_per_cell'] = nbar

        csum = real.csum()
        if pm.comm.rank == 0:
            self.logger.info("painted %d out of %d objects to mesh" %(N,fullsize))
            self.logger.info("mean particles per cell is %g", nbar)
            self.logger.info("sum is %g ", csum)
            self.logger.info("normalized the convention to 1 + delta")

        if normalize:
            if nbar > 0:
                real[...] /= nbar
            else:
                real[...] = 1

        return real

    @property
    def actions(self):
        """
        The actions to apply to the interpolated density field, optionally
        included the compensation correction.
        """
        actions = MeshSource.actions.fget(self)
        if self.compensated:
            actions = self._get_compensation() + actions
        return actions

    def _get_compensation(self):
        """
        Return the compensation function, which corrects for the
        windowing kernel.

        The compensation function is computed as:

        - if ``interlaced = True``:
          - :func:`CompensateCIC` if using CIC window
          - :func:`CompensateTSC` if using TSC window
        - if ``interlaced = False``:
          - :func:`CompensateCICAliasing` if using CIC window
          - :func:`CompensateTSCAliasing` if using TSC window
        """
        if self.interlaced:
            d = {'cic' : self.CompensateCIC,
                 'tsc' : self.CompensateTSC}
        else:
            d = {'cic' : self.CompensateCICAliasing,
                 'tsc' : self.CompensateTSCAliasing}

        if not self.window in d:
            raise ValueError("compensation for window %s is not defined" % self.window)

        filter = d[self.window]

        return [('complex', filter, "circular")]

    @staticmethod
    def CompensateTSC(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the TSC window function in configuration space.

        .. note::
            see equation 18 (with p=3) of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 3
            v = v / tmp
        return v

    @staticmethod
    def CompensateCIC(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the CIC window function in configuration space

        .. note::
            see equation 18 (with p=2) of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 2
            tmp[wi == 0.] = 1.
            v = v / tmp
        return v

    @staticmethod
    def CompensateTSCAliasing(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the TSC window function in configuration space,
        as well as the approximate aliasing correction

        .. note::
            see equation 20 of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            s = numpy.sin(0.5 * wi)**2
            v = v / (1 - s + 2./15 * s**2) ** 0.5
        return v

    @staticmethod
    def CompensateCICAliasing(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the CIC window function in configuration space,
        as well as the approximate aliasing correction

        .. note::
            see equation 20 of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            v = v / (1 - 2. / 3 * numpy.sin(0.5 * wi) ** 2) ** 0.5
        return v



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import kdcount.correlate as correlate
import os
import pytest

# debug logging
setup_logging("debug")

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_data(seed):

    cosmo = cosmology.Planck15
    s = RandomCatalog(1000, seed=seed)

    # ra, dec, z
    s['Redshift'] = s.rng.normal(loc=0.5, scale=0.1, size=s.size)
    s['RA'] = s.rng.uniform(low=110, high=260, size=s.size)
    s['DEC'] = s.rng.uniform(low=-3.6, high=60., size=s.size)

    # position
    s['Position'] = transform.SkyToCartesion(s['RA'], s['DEC'], s['Redshift'], cosmo=cosmo)

    return s

def kcount_paircount(pos1, w1, redges, Nmu, pos2=None, w2=None):
    """
    Verify the paircount algorithm using kdcount
    """
    # get the trees
    tree1 = correlate.points(pos1, boxsize=None, weights=w1)
    if pos2 is None:
        tree2 = tree1
    else:
        tree2 = correlate.points(pos2, boxsize=None, weights=w2)

    # setup the binning
    if Nmu == 0:
        bins = correlate.RBinning(redges)
    else:
        bins = correlate.RmuBinning(redges, Nmu, observer=(0,0,0), mu_min=0., absmu=True)

    # do the paircount
    pc = correlate.paircount(tree1, tree2, bins, np=0, usefast=False, compute_mean_coords=True)

    ravg = pc.mean_centers if Nmu == 0 else pc.mean_centers[0]
    return numpy.nan_to_num(pc.pair_counts), numpy.nan_to_num(ravg), pc.sum1

@MPITest([1, 4])
def test_1d_auto(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # random particles
    source = generate_data(seed=42)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    r = SurveyDataPairCount('1d', source, redges, cosmo, weight='Weight')

    pos = gather_data(source, 'Position')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos, w, redges, 0)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

@MPITest([1, 4])
def test_1d_cross(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # random particles
    source1 = generate_data(seed=42)
    source1['Weight'] = source1.rng.uniform(size=source1.size)
    source2 = generate_data(seed=84)
    source2['Weight'] = source2.rng.uniform(size=source2.size)

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the paircount
    r = SurveyDataPairCount('1d', source1, redges, cosmo, source2=source2)

    pos1 = gather_data(source1, 'Position')
    pos2 = gather_data(source2, 'Position')
    w1 = gather_data(source1, 'Weight')
    w2 = gather_data(source2, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos1, w1, redges, 0, pos2=pos2, w2=w2)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

@MPITest([1, 4])
def test_2d_auto(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # random particles
    source = generate_data(seed=42)
    source['Weight'] = source.rng.uniform(size=source.size)

    # make the bin edges
    redges = numpy.linspace(10, 1000., 10)
    Nmu = 10

    # do the weighted paircount
    r = SurveyDataPairCount('2d', source, redges, cosmo, weight='Weight', Nmu=10)

    pos = gather_data(source, 'Position')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos, w, redges, Nmu)
    #assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])


@MPITest([1, 4])
def test_2d_cross(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # random particles
    source1 = generate_data(seed=42)
    source1['Weight'] = source1.rng.uniform(size=source1.size)
    source2 = generate_data(seed=84)
    source2['Weight'] = source2.rng.uniform(size=source2.size)

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)
    Nmu = 10

    # do the paircount
    r = SurveyDataPairCount('2d', source1, redges, cosmo, source2=source2, Nmu=10)

    pos1 = gather_data(source1, 'Position')
    pos2 = gather_data(source2, 'Position')
    w1 = gather_data(source1, 'Weight')
    w2 = gather_data(source2, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos1, w1, redges, Nmu, pos2=pos2, w2=w2)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

    # test save
    r.save('paircount-test.json')
    r2 = SurveyDataPairCount.load('paircount-test.json')
    assert_array_equal(r.result.data, r2.result.data)

    if comm.rank == 0: os.remove('paircount-test.json')

@MPITest([1])
def test_missing_columns(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # random particles
    source = generate_data(seed=42)

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # missing column
    with pytest.raises(ValueError):
        r = SurveyDataPairCount('1d', source, redges, cosmo, ra='BAD')



from classylss.binding import ClassEngine, Background, Spectra, Perturbs, Primordial, Thermo
from classylss.astropy_compat import AstropyCompat

import numpy
from six import string_types
import os

class Cosmology(object):
    r"""
    A cosmology calculator based on the CLASS binding in :mod:`classylss`.

    It is a collection of all method provided by the CLASS interfaces.

    The individual interfaces can be accessed too, such that
    `c.Spectra.get_transfer` and `c.get_transfer` are identical.

    .. important::

        A default set of units is assumed. Those units are:

        * temperature: :math:`\mathrm{K}`
        * distance: :math:`h^{-1} \mathrm{Mpc}`
        * wavenumber: :math:`h \mathrm{Mpc}^{-1}`
        * power: :math:`h^{-3} \mathrm{Mpc}^3`
        * density: :math:`10^{10} (M_\odot/h) (Mpc/h)^{-3}`
        * neutrino mass: :math:`\mathrm{eV}`
        * time: :math:`\mathrm{Gyr}`
        * H0: :math:`(h^{-1} \mathrm{Mpc}) / (\mathrm{km/s})`

    Notes
    -----
    * The default configuration assumes a flat cosmology, :math:`\Omega_{0,k}=0`.
      Pass ``Omega_k`` in the ``extra`` keyword dictionary to change this value.
    * By default, a cosmological constant is assumed, with its density value
      inferred by the curvature condition.
    * Non-cosmological constant dark energy can be used by specifying the
      ``w0_fld``, ``wa_fld``, and/or ``Omega_fld`` values.
    * The ``sigma8`` attribute can be set to the desired value, and the internal
      value of ``A_s`` will be automatically adjusted.

    Parameters
    ----------
    h : float
        the dimensionaless Hubble parameter
    T_cmb : float
        the temperature of the CMB in Kelvins
    Omega_b : float
        the current baryon density parameter, :math:`\Omega_{b,0}`
    Omega_cdm : float
        the current cold dark matter density parameter, :math:`\Omega_{cdm,0}`
    N_ur : float
        the number of ultra-relativistic (massless neutrino) species; the
        default number is inferred based on the number of massive neutrinos
        via the following logic: if you have respectively 1,2,3 massive
        neutrinos and use the default ``T_ncdm`` value (0.71611 K), designed
        to give m/omega of 93.14 eV, and you wish to have ``N_eff=3.046`` in
        the early universe, then ``N_ur`` is set to 2.0328, 1.0196, 0.00641,
        respectively.
    m_ncdm : list, None
        the masses (in eV) for all massive neutrino species; an empty list
        should  be passed for no massive neutrinso. The default is a single
        massive neutrino with mass of 0.06 eV
    P_k_max : float
        the maximum ``k`` value to compute power spectrum results to, in units
        of :math:`h/Mpc`
    P_z_max : float
        the maximum redshift to compute power spectrum results to
    gauge : str,
        either synchronous or newtonian
    n_s : float
        the tilt of the primordial power spectrum
    nonlinear : bool
        whether to compute nonlinear power spectrum results via HaloFit
    verbose : bool
        whether to turn on the default CLASS logging for all submodules
    **kwargs :
        extra keyword parameters to pass to CLASS; users should be wary
        of configuration options that may conflict with the base set
        of parameters
    """
    # delegate resolve order -- a pun at mro; which in
    # this case introduces the meta class bloat and doesn't solve
    # the issue. We want delayed initialization of interfaces
    # or so-called 'mixins'.
    # easier to just use delegates with a customized getattr.
    # this doesn't work well with automated documentation tools though,
    # unfortunately.

    dro = [AstropyCompat, Thermo, Spectra, Perturbs, Primordial, Background, ClassEngine]
    dro_dict = dict([(n.__name__, n) for n in dro])

    def __init__(self,
            h=0.67556,
            T_cmb=2.7255,
            Omega_b=0.022032/0.67556**2,
            Omega_cdm=0.12038/0.67556**2,
            N_ur=None,
            m_ncdm=[0.06],
            P_k_max=10.,
            P_z_max=100.,
            gauge='synchronous',
            n_s=0.9667,
            nonlinear=False,
            verbose=False,
            **kwargs # additional arguments to pass to CLASS
        ):
        # quickly copy over all arguments --
        # at this point locals only contains the arguments.
        args = dict(locals())

        # store the extra CLASS params
        kwargs = args.pop('kwargs')

        # remove some non-CLASS variables
        args.pop('self')

        # use set state to de-serialize the object.
        self.__setstate__((args,kwargs))

    def __iter__(self):
        """
        Allows dict() to be used on class.
        """
        args = self.args.copy()
        args.update(self.kwargs)
        for k in args:
            yield k, args[k]

    def __dir__(self):
        """ a list of all members from all delegate classes """
        r = []
        # first allow tab completion of delegate names; to help resolve conflicts
        r.extend([n.__name__ for n in self.dro])
        # then allow tab completion of all delegate methods
        for i in reversed(self.dro):
            r.extend(dir(i))
        return sorted(list(set(r)))

    def __getattr__(self, name):
        """
        Find the proper delegate, initialize it, and run the method
        """
        # getting a delegate explicitly, e.g. c.Background
        if name in self.dro_dict:
            iface = self.dro_dict[name]
            if iface not in self.delegates:
                self.delegates[iface] = iface(self.engine)
            return self.delegates[iface]

        # resolving a name from the delegates : c.Om0 => c.Background.Om0
        for iface in self.dro:
            if hasattr(iface, name):
                if iface not in self.delegates:
                    self.delegates[iface] = iface(self.engine)
                d = self.delegates[iface]
                return getattr(d, name)
        else:
            raise AttributeError("Attribute `%s` not found in any of the delegate objects" % name)

    def __getstate__(self):
        return (self.args, self.kwargs)

    @property
    def sigma8(self):
        """
        The amplitude of matter fluctuations at :math:`z=0` in a sphere
        of radius :math:`r = 8 \ h^{-1}\mathrm{Mpc}`.

        This is not an input CLASS parameter, but users can set this parameter
        and the scalar amplitude ``A_s`` will be internally adjusted to
        achieve the desired ``sigma8``.
        """
        return self.Spectra.sigma8

    @sigma8.setter
    def sigma8(self, value):
        if not numpy.isclose(self.sigma8, value):
            set_sigma8(self, value, inplace=True)

    def to_astropy(self):
        """
        Initialize and return a subclass of :class:`astropy.cosmology.FLRW`
        from the :class:`Cosmology` class.

        Returns
        -------
        subclass of :class:`astropy.cosmology.FLRW` :
            the astropy class holding the cosmology values
        """
        import astropy.cosmology as ac
        import astropy.units as au

        is_flat = True
        needs_w0 = False
        needs_wa = False

        pars = {}
        pars['H0'] = 100*self.h
        pars['Om0'] = self.Omega0_b + self.Omega0_cdm # exclude massive neutrinos to better match astropy
        pars['Tcmb0'] = self.Tcmb0
        pars['Neff'] = self.Neff
        pars['Ob0'] = self.Ob0

        if self.has_massive_nu:

            # all massless by default
            m_nu = numpy.zeros(int(numpy.floor(self.Neff)))

            # then add massive species
            m_nu[:len(self.m_ncdm)] = self.m_ncdm[:]
            pars['m_nu'] = au.Quantity(m_nu, au.eV)

        if self.Ok0 != 0.:
            pars['Ode0'] = self.Ode0
            is_flat = False

        if self.wa_fld != 0:
            pars['wa'] = self.wa_fld
            pars['w0'] = self.wa_fld
            needs_wa = True

        if self.w0_fld != -1:
            pars['w0'] = self.w0_fld
            needs_w0 = True

        # determine class to return
        prefix = "" if not is_flat else "Flat"
        if needs_wa:
            cls = prefix + "w0waCDM"
        elif needs_w0:
            cls = prefix + "wCDM"
        else:
            cls = prefix + "LambdaCDM"
        cls = getattr(ac, cls)

        print(cls)
        return cls(**pars)

    @classmethod
    def from_astropy(kls, cosmo, **kwargs):
        """
        Initialize and return a :class:`Cosmology` object from a subclass of
        :class:`astropy.cosmology.FLRW`.

        Parameters
        ----------
        cosmo : subclass of :class:`astropy.cosmology.FLRW`.
            the astropy cosmology instance
        **kwargs :
            extra keyword parameters to pass when initializing

        Returns
        -------
        :class:`Cosmology` :
            the initialized cosmology object
        """
        args = astropy_to_dict(cosmo)
        args.update(kwargs)
        return Cosmology(**args)

    @classmethod
    def from_file(cls, filename, **kwargs):
        """
        Initialize a :class:`Cosmology` object from the CLASS parameter file

        Parameters
        ----------
        filename : str
            the name of the parameter file to read
        """
        from classylss import load_ini

        # extract dictionary of parameters from the file
        pars = load_ini(filename)
        pars.update(**kwargs)

        # initialize the engine as the backup delegate.
        toret = object.__new__(cls)
        toret.engine = ClassEngine(pars)
        toret.delegates = {ClassEngine: toret.engine}

        # reconstruct the correct __init__ params
        args, kwargs = sanitize_class_params(toret, pars)

        toret.args = args
        toret.kwargs = kwargs
        return toret

    def __setstate__(self, state):

        # remember for serialization
        self.args, self.kwargs = state

        # verify and set defaults
        pars = verify_parameters(self.args, self.kwargs)

        # initialize the engine as the backup delegate.
        self.engine = ClassEngine(pars)
        self.delegates = {ClassEngine: self.engine}

    def clone(self, **kwargs):
        """
        Create a new cosmology based on modification of self, with the
        input keyword parameters changed.

        Parameters
        ----------
        **kwargs :
            keyword parameters to adjust

        Returns
        -------
        :class:`Cosmology`
            a copy of self, with the input ``kwargs`` adjusted
        """
        # initialize a new object (so we have sanitized args/kwargs)
        new = Cosmology(**kwargs)

        # the named keywords
        args = self.args.copy()
        args.update(new.args)

        # the extra keywords
        kwargs = self.kwargs.copy()
        kwargs.update(new.kwargs)
        args.update(kwargs)

        return Cosmology(**args)


def astropy_to_dict(cosmo):
    """
    Convert an astropy cosmology object to a dictionary of parameters
    suitable for initializing a Cosmology object.
    """
    from astropy import cosmology, units

    pars = {}
    pars['h'] = cosmo.h
    pars['T_cmb'] = cosmo.Tcmb0.value
    if cosmo.Ob0 is not None:
        pars['Omega_b'] = cosmo.Ob0
    else:
        raise ValueError("please specify a value 'Ob0' ")
    pars['Omega_cdm'] = cosmo.Om0 - cosmo.Ob0 # should be okay for now

    # handle massive neutrinos
    if cosmo.has_massive_nu:

        # convert to eV
        m_nu = cosmo.m_nu
        if hasattr(m_nu, 'unit') and m_nu.unit != units.eV:
            m_nu = m_nu.to(units.eV)
        else:
            m_nu = units.eV * m_nu
        # from CLASS notes:
        # one more remark: if you have respectively 1,2,3 massive neutrinos,
        # if you stick to the default value pm equal to 0.71611, designed to give m/omega of
        # 93.14 eV, and if you want to use N_ur to get N_eff equal to 3.046 in the early universe,
        # then you should pass here respectively 2.0328,1.0196,0.00641
        N_ur = [2.0328, 1.0196, 0.00641]
        N_massive = (m_nu > 0.).sum()
        pars['N_ur'] = (cosmo.Neff/3.046) * N_ur[N_massive-1]

        pars['m_ncdm'] = [k.value for k in sorted(m_nu[m_nu > 0.], reverse=True)]
    else:
        pars['m_ncdm'] = []
        pars['N_ur'] = cosmo.Neff

    # specify the curvature
    pars['Omega_k'] = cosmo.Ok0

    # handle dark energy
    if isinstance(cosmo, cosmology.LambdaCDM):
        pass
    elif isinstance(cosmo, cosmology.wCDM):
        pars['w0_fld'] = cosmo.w0
        pars['wa_fld'] = 0.
        pars['Omega_Lambda'] = 0. # use Omega_fld
    elif isinstance(cosmo, cosmology.w0waCDM):
        pars['w0_fld'] = cosmo.w0
        pars['wa_fld'] = cosmo.wa
        pars['Omega_Lambda'] = 0. # use Omega_fld
    else:
        cls = cosmo.__class__.__name__
        valid = ["LambdaCDM", "wCDM", "w0waCDM"]
        msg = "dark energy equation of state not recognized for class '%s'; " %cls
        msg += "valid classes: %s" %str(valid)
        raise ValueError(msg)

    return pars

def verify_parameters(args, extra):
    """
    Verify the input parameters to a :class:`Cosmology` object and
    set various default values.
    """
    # check for conflicts
    for par in CONFLICTS:
        for p in CONFLICTS[par]:
            if p in extra:
                raise ValueError("input parameter conflict; use '%s', not '%s'" %(par, p))

    pars = {}
    pars.update(args)
    pars.update(extra)

    # set some default parameters
    pars.setdefault('output', "vTk dTk mPk")
    pars.setdefault('extra metric transfer functions', 'y')

    # no massive neutrinos
    if pars.get('m_ncdm', None) is None:
        pars['m_ncdm'] = []

    # a single massive neutrino
    if numpy.isscalar(pars['m_ncdm']):
        pars['m_ncdm'] = [pars['m_ncdm']]

    # needs to be a list
    if not isinstance(pars['m_ncdm'], (list,numpy.ndarray)):
        raise TypeError("``m_ncdm`` should be a list of mass values in eV")

    # check gauge
    if pars.get('gauge', 'synchronous') not in ['synchronous', 'newtonian']:
        raise ValueError("'gauge' should be 'synchronous' or 'newtonian'")

    for m in pars['m_ncdm']:
        if m == 0:
            raise ValueError("A zero mass is specified in the non-cold dark matter list. "
                             "This is not needed, as we automatically set N_ur based on "
                             "the number of entries in m_ncdm such that Neff = 3.046.")

    # remove None's -- use CLASS default
    for key in list(pars.keys()):
        if pars[key] is None: pars.pop(key)

    # set cosmological constant to zero if we got fluid w0/wa
    if 'w0_fld' in pars or 'wa_fld' in pars:
        if pars.get('Omega_Lambda', 0) > 0:
            raise ValueError(("non-zero fOmega_Lambda (cosmological constant) specified as "
                             "well as fluid w0/wa; use Omega_fld instead"))
        pars['Omega_Lambda'] = 0.

    # turn on verbosity
    verbose = pars.pop('verbose', False)
    if verbose:
        for par in ['input', 'background', 'thermodynamics', 'perturbations',
                    'transfer', 'primordial', 'spectra', 'nonlinear', 'lensing']:
            name = par + '_verbose'
            if name not in pars: pars[name] = 1

    # maximum k value
    if 'P_k_max_h/Mpc' not in pars:
        pars['P_k_max_h/Mpc'] = pars.pop('P_k_max', 10.)

    # maximum redshift
    if 'z_max_pk' not in pars:
        pars['z_max_pk'] = pars.pop('P_z_max', 100.)

    # nonlinear power?
    if 'non linear' not in pars:
        if pars.pop('nonlinear', False):
            pars['non linear'] = 'halofit'

    # number of massive neutrino species
    pars['N_ncdm'] = len(pars['m_ncdm'])

    # m_ncdm only needed if we have massive neutrinos
    if not pars['N_ncdm']: pars.pop('m_ncdm')

    # from CLASS notes:
    # one more remark: if you have respectively 1,2,3 massive neutrinos,
    # if you stick to the default value pm equal to 0.71611, designed to give m/omega of
    # 93.14 eV, and if you want to use N_ur to get N_eff equal to 3.046 in the early universe,
    # then you should pass here respectively 2.0328,1.0196,0.00641
    N_ur_table = [3.046, 2.0328, 1.0196, 0.00641]
    if 'N_ur' not in pars:
        pars['N_ur'] = N_ur_table[pars['N_ncdm']]

    return pars


def set_sigma8(cosmo, sigma8, inplace=False):
    """
    Return a clone of the input Cosmology object, with the ``sigma8`` value
    set to the specified value.

    Parameters
    ----------
    cosmo : Cosmology
        the input cosmology object
    sigma8 : float
        the desired sigma8 value
    inplace : bool, optional
        if ``True``, update sigma8 of the input ``cosmo`` object, else return
        a new Cosmology object
    """
    # the new scalar amplitude A_s
    A_s = cosmo.A_s * (sigma8/cosmo.sigma8)**2

    # the extra keywords
    kwargs = cosmo.kwargs.copy()

    # set the desired A_s and remove conflicting parameters
    kwargs['A_s'] = A_s
    kwargs.pop('ln10^{10}A_s', None)

    if inplace:
        cosmo.__setstate__((cosmo.args, kwargs))
    else:
        # add the name keywords
        kwargs.update(cosmo.args)

        # new cosmo clone
        cosmo = Cosmology(**kwargs)

    return cosmo

def sanitize_class_params(cosmo, pars):
    """
    Given a dictionary of CLASS parameters, construct the ``args``
    dict and ``kwargs`` dict that can used to initialize a
    Cosmology class, accounting for any possible conflicts.

    The ``args`` dict holds the main (named) __init__ keywords, and the
    ``kwargs`` holds all of the extra keywords.
    """
    args = {}
    kwargs = pars.copy()

    # loop over all parameters
    for name in list(kwargs.keys()):

        # parameter is a main parameter
        if name in CONFLICTS:
            kwargs.pop(name)
            alias = ALIASES.get(name, name) # check for attribute alias
            args[name] = getattr(cosmo, alias)
        else:
            # check if parameter conflicts with main parameter
            for c in CONFLICTS:
                if name in CONFLICTS[c]:
                    kwargs.pop(name)
                    alias = ALIASES.get(c, c)
                    args[c] = getattr(cosmo, alias)

    # set all named keywords that do not have parameter conflicts
    args['gauge'] = cosmo.gauge

    return args, kwargs

# dict mapping input CLASS params to the Cosmology attribute name
ALIASES = {'Omega_b': 'Omega0_b', 'Omega_cdm':'Omega0_cdm', 'T_cmb':'T0_cmb',
           'P_k_max': 'k_max_for_pk'}

# dict that defines input parameters that conflict with each other
CONFLICTS = {'h': ['H0', '100*theta_s'],
             'T_cmb': ['Omega_g', 'omega_g'],
             'Omega_b': ['omega_b'],
             'N_ur': ['Omega_ur', 'omega_ur'],
             'Omega_cdm': ['omega_cdm'],
             'm_ncdm': ['Omega_ncdm', 'omega_ncdm'],
             'P_k_max': ['P_k_max_h/Mpc', 'P_k_max_1/Mpc'],
             'P_z_max': ['z_max_pk'],
             'nonlinear' : ['non linear']
            }



from nbodykit.source.catalogmesh.species import MultipleSpeciesCatalogMesh
import logging

class FKPCatalogMesh(MultipleSpeciesCatalogMesh):
    """
    A subclass of
    :class:`~nbodykit.source.catalogmesh.species.MultipleSpeciesCatalogMesh`
    designed to paint a :class:`~nbodykit.source.catalog.fkp.FKPCatalog` to
    a mesh.

    The multiple species here are ``data`` and ``randoms`` CatalogSource
    objects, where ``randoms`` is a catalog of randomly distributed objects
    with no instrinsic clustering that defines the survey volume.

    Internally, all of the columns in ``data`` and ``randoms`` are stored,
    with names prefixed by ``data/`` or ``randoms/``.

    Parameters
    ----------
    source : CatalogSource
        the input catalog that we wish to interpolate to a mesh
    BoxSize :
        the size of the box
    Nmesh : int, 3-vector
        the number of cells per mesh side
    dtype : str
        the data type of the values stored on mesh
    selection : str
        column in ``source`` that selects the subset of particles to grid
        to the mesh
    comp_weight : str
        the completeness weight column name
    fkp_weight : str
        the FKP weight column name
    nbar : str
        the n(z) column name
    position : str, optional
        column in ``source`` specifying the position coordinates; default
        is ``Position``
    """
    logger = logging.getLogger('FKPCatalogMesh')

    def __init__(self, source, BoxSize, Nmesh, dtype, selection,
                    comp_weight, fkp_weight, nbar, position='Position'):

        from nbodykit.source.catalog import FKPCatalog
        if not isinstance(source, FKPCatalog):
            raise TypeError("the input source for FKPCatalogMesh must be a FKPCatalog")

        self.source  = source
        self.comp_weight = comp_weight
        self.fkp_weight = fkp_weight
        self.nbar = nbar

        # store the input position column
        self._position = position

        # init the base
        # NOTE: FKP must paint the Position recentered to [-L/2, L/2]
        # so we store that in an internal column "_RecenteredPosition"
        MultipleSpeciesCatalogMesh.__init__(self, source, BoxSize, Nmesh, dtype,
                                            '_TotalWeight', 'Value', selection,
                                            position='_RecenteredPosition')


    def to_real_field(self):
        r"""
        Paint the FKP density field, returning a ``RealField``.

        Given the ``data`` and ``randoms`` catalogs, this paints:

        .. math::

            F(x) = w_\mathrm{fkp}(x) * [w_\mathrm{comp}(x)*n_\mathrm{data}(x) -
                        \alpha * w_\mathrm{comp}(x)*n_\mathrm{randoms}(x)]


        This computes the following meta-data attributes in the process of
        painting, returned in the :attr:`attrs` attributes of the returned
        RealField object:

        - randoms.W, data.W :
            the weighted sum of randoms and data objects; see
            :func:`weighted_total`
        - alpha : float
            the ratio of ``data.W`` to ``randoms.W``
        - randoms.norm, data.norm : float
            the power spectrum normalization; see :func:`normalization`
        - randoms.shotnoise, data.shotnoise: float
            the shot noise for each sample; see :func:`shotnoise`
        - shotnoise : float
            the total shot noise, equal to the sum of ``randoms.shotnoise``
            and ``data.shotnoise``
        - randoms.num_per_cell, data.num_per_cell : float
            the mean number of weighted objects per cell for each sample
        - num_per_cell : float
            the mean number of weighted objects per cell

        For further details on the meta-data, see
        :ref:`the documentation <fkp-meta-data>`.

        Returns
        -------
        :class:`~pmesh.pm.RealField` :
            the field object holding the FKP density field in real space
        """
        # add necessary FKP columns first
        # NOTE: these are internal columns and will be deleted so successive
        # calls to this function remain valid
        for i, name in enumerate(self.source.species):

            # a total weight for the mesh is completeness weight x FKP weight
            self[name+'/_TotalWeight'] = self[name+'/'+self.comp_weight] * self[name+'/'+self.fkp_weight]

            # position on the mesh is re-centered to [-BoxSize/2, BoxSize/2]
            self[name+'/_RecenteredPosition'] = self[name+'/'+self._position] - self.source.attrs['BoxCenter']

        attrs = {}

        # determine alpha, the weighted number ratio
        for name in self.source.species:
            attrs[name+'.W'] = self.weighted_total(name)
        attrs['alpha'] = attrs['data.W'] / attrs['randoms.W']

        # randoms get an additional weight of -alpha
        self['randoms/_TotalWeight'] *= -1.0 * attrs['alpha']

        # paint w_data*n_data - alpha*w_randoms*n_randoms
        real = MultipleSpeciesCatalogMesh.to_real_field(self, normalize=False)

        # the rest of the meta-data
        for name in self.source.species:
            attrs[name + '.norm'] = self.normalization(name)
            attrs[name + '.shotnoise'] = self.shotnoise(name)

        # finish the statistics
        attrs['randoms.norm'] *= attrs['alpha']
        attrs['randoms.shotnoise'] *= attrs['alpha']**2 / attrs['randoms.norm']
        attrs['data.shotnoise'] /=  attrs['randoms.norm']
        attrs['shotnoise'] = attrs['data.shotnoise'] + attrs['randoms.shotnoise']

        # divide by volume per cell to go from number to number density
        vol_per_cell = (self.pm.BoxSize/self.pm.Nmesh).prod()
        real[:] /= vol_per_cell

        # update the meta-data
        real.attrs.update(attrs)

        # delete the INTERNAL columns we added while painting
        for i, name in enumerate(self.source.species):
            del self[name+'/_TotalWeight']
            del self[name+'/_RecenteredPosition']

        return real

    def normalization(self, name):
        r"""
        Compute the power spectrum normalization, using either the
        ``data`` or ``randoms`` source.

        This computes:

        .. math::

            A = \sum \bar{n} w_\mathrm{comp} w_\mathrm{fkp}^2

        References
        ----------
        see Eqs. 13,14 of Beutler et al. 2014, "The clustering of galaxies in the
        SDSS-III Baryon Oscillation Spectroscopic Survey: testing gravity with redshift
        space distortions using the power spectrum multipoles"
        """
        sel = self.source.compute(self[name+'/'+self.selection])

        nbar        = self[name+'/'+self.nbar][sel]
        comp_weight = self[name+'/'+self.comp_weight][sel]
        fkp_weight  = self[name+'/'+self.fkp_weight][sel]
        A           = nbar*comp_weight*fkp_weight**2

        A = self.source.compute(A.sum())
        return self.comm.allreduce(A)

    def shotnoise(self, name):
        r"""
        Compute the power spectrum shot noise, using either the
        ``data`` or ``randoms`` source.

        This computes:

        .. math::

            S = \sum (w_\mathrm{comp} w_\mathrm{fkp})^2

        References
        ----------
        see Eq. 15 of Beutler et al. 2014, "The clustering of galaxies in the
        SDSS-III Baryon Oscillation Spectroscopic Survey: testing gravity with redshift
        space distortions using the power spectrum multipoles"
        """
        sel = self.source.compute(self[name+'/'+self.selection])

        comp_weight = self[name+'/'+self.comp_weight][sel]
        fkp_weight  = self[name+'/'+self.fkp_weight][sel]
        S           = (comp_weight*fkp_weight)**2

        S = self.source.compute(S.sum())
        return self.comm.allreduce(S)

    def weighted_total(self, name):
        r"""
        Compute the weighted total number of objects, using either the
        ``data`` or ``randoms`` source:

        This is the sum of the completeness weights:

        .. math::

            W = \sum w_\mathrm{comp}
        """
        sel = self.source.compute(self[name+'/'+self.selection])
        comp_weight = self[name+'/'+self.comp_weight][sel]

        wsum = self.source.compute(comp_weight.sum())
        return self.comm.allreduce(wsum)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_fof(comm):
    cosmo = cosmology.Planck15

    CurrentMPIComm.set(comm)

    # lognormal particles
    source = LogNormalCatalog(Plin=cosmology.LinearPower(cosmo, 0.55),
                nbar=3e-3, BoxSize=512., Nmesh=128, seed=42)

    # compute P(k,mu) and multipoles
    fof = FOF(source, linking_length=0.2, nmin=20)
    source['Density'] = KDDensity(source, margin=1).density

    # save the halos
    peaks = fof.find_features(peakcolumn='Density')
    peaks.save("FOF-%d" % comm.size, ['CMPosition', 'CMVelocity', 'Length', 'PeakPosition', 'PeakVelocity'])



from nbodykit.base.catalog import CatalogSource, column
from nbodykit import transform
import numpy

class HaloCatalog(CatalogSource):
    """
    A wrapper CatalogSource of halo objects to interface nicely with
    :class:`halotools.sim_manager.UserSuppliedHaloCatalog`.

    Parameters
    ----------
    source : CatalogSource
        the source holding the particles to be interpreted as halos
    cosmo : :class:`~nbodykit.cosmology.core.Cosmology`
        the cosmology instance;
    redshift : float
        the redshift of the halo catalog
    mdef : str, optional
        string specifying mass definition, used for computing default
        halo radii and concentration; should be 'vir' or 'XXXc' or
        'XXXm' where 'XXX' is an int specifying the overdensity
    mass : str, optional
        the column name specifying the mass of each halo
    position : str, optional
        the column name specifying the position of each halo
    velocity : str, optional
        the column name specifying the velocity of each halo
    """
    def __init__(self, source, cosmo, redshift, mdef='vir',
                 mass='Mass', position='Position', velocity='Velocity'):

        # make sure all of the columns are there
        for name, col in zip(['mass', 'position', 'velocity'], [mass, position, velocity]):
            if col is None:
                raise ValueError("the %s column cannot be None in HaloCatalog" %name)
            if col not in source:
                raise ValueError("input source is missing the %s column; '%s' does not exist" %(name, col))

        comm = source.comm
        self._source = source
        self.cosmo = cosmo

        # get the attrs from the source
        self.attrs.update(source.attrs)

        # and save the parameters
        self.attrs['redshift'] = redshift
        self.attrs['mass']     = mass
        self.attrs['velocity'] = velocity
        self.attrs['position'] = position
        self.attrs['mdef']     = mdef

        # init the base class
        CatalogSource.__init__(self, comm=comm, use_cache=False)

    @property
    def size(self):
        return self._source.size

    @column
    def Mass(self):
        """
        The halo mass column, assumed to be in units of :math:`M_\odot/h`.
        """
        return self.make_column(self._source[self.attrs['mass']])

    @column
    def Position(self):
        """
        The halo position column, assumed to be in units of :math:`\mathrm{Mpc}/h`.
        """
        return self.make_column(self._source[self.attrs['position']])

    @column
    def Velocity(self):
        """
        The halo velocity column, assumed to be in units of km/s.
        """
        return self.make_column(self._source[self.attrs['velocity']])

    @column
    def VelocityOffset(self):
        """
        The redshift-space distance offset due to the velocity in units of
        distance. The assumed units are :math:`\mathrm{Mpc}/h`.

        This multiplies ``Velocity`` by :math:`1 / (a 100 E(z)) = 1 / (a H(z)/h)`.
        """
        z = self.attrs['redshift']
        rsd_factor = (1+z) / (100*self.cosmo.efunc(z))
        return self['Velocity'] * rsd_factor

    @column
    def Concentration(self):
        """
        The halo concentration, computed using :func:`nbodykit.transform.HaloConcentration`.

        This uses the analytic formulas for concentration from
        `Dutton and Maccio 2014 <https://arxiv.org/abs/1402.7073>`_.
        """
        z = self.attrs['redshift']
        mdef = self.attrs['mdef']
        return transform.HaloConcentration(self['Mass'], self.cosmo, z, mdef=mdef)

    @column
    def Radius(self):
        """
        The halo radius, computed using :func:`nbodykit.transform.HaloRadius`.

        Assumed units of :math:`\mathrm{Mpc}/h`.
        """
        z = self.attrs['redshift']
        mdef = self.attrs['mdef']
        return transform.HaloRadius(self['Mass'], self.cosmo, z, mdef=mdef)

    def to_halotools(self, BoxSize=None, selection='Selection'):
        """
        Return the CatalogSource as a :class:`halotools.sim_manager.UserSuppliedHaloCatalog`.

        The Halotools catalog only holds the local data, although halos are
        labeled via the ``halo_id`` column using the global index.

        Parameters
        ----------
        BoxSize : float, array_like, optional
            the size of the box; note that anisotropic boxes are currently
            not supported by halotools
        selection : str, optional
            the name of the column to slice the data on before converting
            to a halotools catalog

        Returns
        -------
        cat : :class:`halotools.sim_manager.UserSuppliedHaloCatalog`
            the Halotools halo catalog, storing the local halo data
        """
        from halotools.sim_manager import UserSuppliedHaloCatalog
        from halotools.empirical_models import model_defaults

        # make sure we have at least one halo
        if self.csize == 0:
            raise ValueError("cannot create halotools UserSuppliedHaloCatalog; catalog is empty")

        # make sure we have a BoxSize
        if BoxSize is None:
            BoxSize = self.attrs.get('BoxSize', None)
        if BoxSize is None:
            raise ValueError("please specify a 'BoxSize' to convert to a halotools catalog")

        # anisotropic boxes not yet supported
        if not numpy.isscalar(BoxSize):
            BoxSize = numpy.asarray(BoxSize)
            if not (BoxSize == BoxSize[0]).all():
                raise ValueError("halotools does not currently support anisotropic boxes; see astropy/halotools#641")
            else:
                BoxSize = BoxSize[0]

        assert selection in self, "'%s' selection column is not valid" %selection

        # compute the columns
        sel = self.compute(self[selection])
        cols = ['Position', 'Velocity', 'Mass', 'Radius', 'Concentration']
        cols = self.compute(*[self[col][sel] for col in cols])
        Position, Velocity, Mass, Radius, Concen = [col for col in cols]

        # names of the mass and radius fields, based on mass def
        mkey = model_defaults.get_halo_mass_key(self.attrs['mdef'])
        rkey = model_defaults.get_halo_boundary_key(self.attrs['mdef'])

        # global halo ids (across all ranks)
        sizes = self.comm.allgather(self.size)
        start = sum(sizes[:self.comm.rank])
        halo_id = numpy.arange(start, start+self.size, dtype='i8')[sel]

        # data columns
        kws                  = {}
        kws['halo_x']        = Position[:,0]
        kws['halo_y']        = Position[:,1]
        kws['halo_z']        = Position[:,2]
        kws['halo_vx']       = Velocity[:,0]
        kws['halo_vy']       = Velocity[:,1]
        kws['halo_vz']       = Velocity[:,2]
        kws[mkey]            = Mass
        kws[rkey]            = Radius
        kws['halo_nfw_conc'] = Concen
        kws['halo_id']       = halo_id
        kws['halo_upid']     = numpy.zeros(len(Position)) - 1
        kws['halo_local_id'] = numpy.arange(0, self.size, dtype='i8')[sel]

        # add metadata too
        kws['cosmology']     = self.cosmo
        kws['redshift']      = self.attrs['redshift']
        kws['Lbox']          = BoxSize
        kws['particle_mass'] = self.attrs.get('particle_mass', 1.0)
        kws['mdef']          = self.attrs['mdef']

        return UserSuppliedHaloCatalog(**kws)



from six import add_metaclass, string_types
from nbodykit.transform import ConstantArray

import abc
import numpy
import logging
import warnings
import dask.array as da

# default size of Cache for CatalogSource arrays
CACHE_SIZE = 1e9

class ColumnAccessor(da.Array):
    """
    Provides access to a Column from a Catalog

    This is a thin subclass of :class:`dask.array.Array` to
    provide a reference to the catalog object,
    an additional ``attrs`` attribute (for recording the
    reproducible meta-data), and some pretty print support.

    Due to particularity of :mod:`dask`, any transformation
    that is not explicitly in-place will return
    a :class:`dask.array.Array`, and losing the pointer to
    the original catalog and the meta data attrs.
    """
    def __new__(cls, catalog, daskarray):
        self = da.Array.__new__(ColumnAccessor,
                daskarray.dask,
                daskarray.name,
                daskarray.chunks,
                daskarray.dtype,
                daskarray.shape)
        self.catalog = catalog
        self.attrs = {}
        return self

    def as_daskarray(self):
        return da.Array(
                self.dask,
                self.name,
                self.chunks,
                self.dtype,
                self.shape)

    def compute(self):
        return self.catalog.compute(self)

    def __str__(self):
        r = da.Array.__str__(self)
        if len(self) > 0:
            r = r + " first : %s" % str(self[0].compute())
        if len(self) > 1:
            r = r + " last: %s" % str(self[-1].compute())
        return r

def column(name=None):
    """
    Decorator that defines a function as a column in a CatalogSource
    """
    def decorator(getter):
        getter.column_name = name
        return getter

    if hasattr(name, '__call__'):
        # a single callable is provided
        getter = name
        name = getter.__name__
        return decorator(getter)
    else:
        return decorator


def find_columns(cls):
    """
    Find all hard-coded column names associated with the input class

    Returns
    -------
    hardcolumns : set
        a set of the names of all hard-coded columns for the
        input class ``cls``
    """
    hardcolumns = []

    # search through the class dict for columns
    for key, value in cls.__dict__.items():
         if hasattr(value, 'column_name'):
            hardcolumns.append(value.column_name)

    # recursively search the base classes, too
    for base in cls.__bases__:
        hardcolumns += find_columns(base)

    return list(sorted(set(hardcolumns)))


def find_column(cls, name):
    """
    Find a specific column ``name`` of an input class, or raise
    an exception if it does not exist

    Returns
    -------
    column : callable
        the callable that returns the column data
    """
    # first search through class
    for key, value in cls.__dict__.items():
        if not hasattr(value, 'column_name'): continue
        if value.column_name == name: return value

    for base in cls.__bases__:
        try: return find_column(base, name)
        except: pass

    args = (name, str(cls))
    raise AttributeError("unable to find column '%s' for '%s'" % args)


class CatalogSourceBase(object):
    """
    An abstract base class representing a catalog of discrete particles.

    This objects behaves like a structured numpy array -- it must have a
    well-defined size when initialized. The ``size`` here represents the
    number of particles in the source on the local rank.

    Subclasses of this class must define a ``size`` attribute.

    The information about each particle is stored as a series of
    columns in the format of dask arrays. These columns can be accessed
    in a dict-like fashion.
    """
    logger = logging.getLogger('CatalogSourceBase')

    @staticmethod
    def make_column(array):
        """
        Utility function to convert a numpy array to a :class:`dask.array.Array`.
        """
        if isinstance(array, da.Array):
            return array
        elif isinstance(array, ColumnAccessor):
            # important to get the accessor as a dask array to avoid circular
            # references
            return array.as_daskarray()
        else:
            return da.from_array(array, chunks=100000)

    def __init__(self, comm, use_cache=False):

        # ensure self.comm is set, though usually already set by the child.
        self.comm = comm

        # initialize a cache
        self.use_cache = use_cache

        # user-provided overrides for columns
        self._overrides = {}

    def __iter__(self):
        return iter(self.columns)

    def __contains__(self, col):
        return col in self.columns

    def __getitem__(self, sel):
        """
        The following types of indexing are supported:

        #.  strings specifying a column in the CatalogSource; returns
            a dask array holding the column data
        #.  boolean arrays specifying a slice of the CatalogSource;
            returns a CatalogCopy holding only the revelant slice
        #.  slice object specifying which particles to select
        #.  list of strings specifying column names; returns a CatalogCopy
            holding only the selected columnss
        """
        # handle boolean array slices
        if not isinstance(sel, string_types):

            # convert slices to boolean arrays
            if isinstance(sel, (slice, list)):

                # select a subset of list of string column names
                if isinstance(sel, list) and all(isinstance(ss, string_types) for ss in sel):
                    invalid = set(sel) - set(self.columns)
                    if len(invalid):
                        msg = "cannot select subset of columns from "
                        msg += "CatalogSource due to invalid columns: %s" % str(invalid)
                        raise KeyError(msg)

                    # return a CatalogSource only holding the selected columns
                    subset_data = {col:self[col] for col in sel}
                    toret = CatalogCopy(self.size, self.comm, use_cache=self.use_cache, **subset_data)
                    toret.attrs.update(self.attrs)
                    return toret

                # list must be all integers
                if isinstance(sel, list) and not numpy.array(sel).dtype == numpy.integer:
                    raise KeyError("array like indexing via a list should be a list of integers")

                index = numpy.zeros(self.size, dtype='?')
                index[sel] = True; sel = index

            # do the slicing
            if not numpy.isscalar(sel):
                return get_catalog_subset(self, sel)
            else:
                raise KeyError("strings and boolean arrays are the only supported indexing methods")

        # get the right column
        if sel in self._overrides:
            r = self._overrides[sel]
        elif sel in self.hardcolumns:
            r = self.get_hardcolumn(sel)
        else:
            raise KeyError("column `%s` is not defined in this source; " %sel + \
                            "try adding column via `source[column] = data`")

        # evaluate callables if we need to
        if callable(r): r = r()

        r = ColumnAccessor(self, r)

        return r

    def __setitem__(self, col, value):
        """
        Add new columns to the CatalogSource, overriding any existing columns
        with the name ``col``.
        """
        self._overrides[col] = self.make_column(value)

    def __delitem__(self, col):
        """
        Delete a column; cannot delete a "hard-coded" column
        """
        if col not in self.columns:
            raise ValueError("no such column, cannot delete it")

        if col in self.hardcolumns:
            raise ValueError("cannot delete a hard-coded column")

        if col in self._overrides:
            del self._overrides[col]
            return

        raise ValueError("unable to delete column '%s' from CatalogSource" %col)


    @property
    def use_cache(self):
        """
        If set to ``True``, use the built-in caching features of ``dask``
        to cache data in memory.
        """
        return self._use_cache

    @use_cache.setter
    def use_cache(self, val):
        """
        Initialize a Cache object of size set by ``CACHE_SIZE``, which
        is 1 GB by default.
        """
        if val:
            try:
                from dask.cache import Cache
                if not hasattr(self, '_cache'):
                    self._cache = Cache(CACHE_SIZE)
            except ImportError:
                warnings.warn("caching of CatalogSource requires ``cachey`` module; turning cache off")
        else:
            if hasattr(self, '_cache'): delattr(self, '_cache')
        self._use_cache = val

    @property
    def hardcolumns(self):
        """
        A list of the hard-coded columns in the CatalogSource.

        These columns are usually member functions marked by @column decorator.
        Subclasses may override this method and use :func:`get_hardcolumn` to
        bypass the decorator logic.
        """
        try:
            self._hardcolumns
        except AttributeError:
            self._hardcolumns = find_columns(self.__class__)
        return self._hardcolumns

    @property
    def columns(self):
        """
        All columns in the CatalogSource, including those hard-coded into
        the class's defintion and override columns provided by the user.
        """
        hcolumns  = list(self.hardcolumns)
        overrides = list(self._overrides)
        return sorted(set(hcolumns + overrides))

    @property
    def attrs(self):
        """
        A dictionary storing relevant meta-data about the CatalogSource.
        """
        try:
            return self._attrs
        except AttributeError:
            self._attrs = {}
            return self._attrs



    def get_hardcolumn(self, col):
        """
        Construct and return a hard-coded column.

        These are usually produced by calling member functions marked by the
        @column decorator.

        Subclasses may override this method and the hardcolumns attribute to
        bypass the decorator logic.
        """
        return find_column(self.__class__, col)(self)

    def compute(self, *args, **kwargs):
        """
        Our version of :func:`dask.compute` that computes
        multiple delayed dask collections at once.

        This should be called on the return value of :func:`read`
        to converts any dask arrays to numpy arrays.

        If :attr:`use_cache` is ``True``, this internally caches data, using
        dask's built-in cache features.

        Parameters
        -----------
        args : object
            Any number of objects. If the object is a dask
            collection, it's computed and the result is returned.
            Otherwise it's passed through unchanged.

        Notes
        -----
        The dask default optimizer induces too many (unnecesarry)
        IO calls -- we turn this off feature off by default. Eventually we
        want our own optimizer probably.
        """
        import dask

        # do not optimize graph (can lead to slower optimizations)
        kwargs.setdefault('optimize_graph', False)

        # use a cache?
        if self.use_cache and hasattr(self, '_cache'):
            with self._cache:
                toret = dask.compute(*args, **kwargs)
        else:
            toret = dask.compute(*args, **kwargs)

        # do not return tuples of length one
        if len(toret) == 1: toret = toret[0]
        return toret

    def save(self, output, columns, datasets=None, header='Header'):
        """
        Save the CatalogSource to a :class:`bigfile.BigFile`.

        Only the selected columns are saved and :attr:`attrs` are saved in
        ``header``. The attrs of columns are stored in the datasets.

        Parameters
        ----------
        output : str
            the name of the file to write to
        columns : list of str
            the names of the columns to save in the file
        datasets : list of str, optional
            names for the data set where each column is stored; defaults to
            the name of the column
        header : str, optional
            the name of the data set holding the header information, where
            :attr:`attrs` is stored
        """
        import bigfile
        import json
        from nbodykit.utils import JSONEncoder

        if datasets is None: datasets = columns
        if len(datasets) != len(columns):
            raise ValueError("`datasets` must have the same length as `columns`")

        with bigfile.BigFileMPI(comm=self.comm, filename=output, create=True) as ff:
            try:
                bb = ff.open(header)
            except:
                bb = ff.create(header)
            with bb :
                for key in self.attrs:
                    try:
                        bb.attrs[key] = self.attrs[key]
                    except ValueError:
                        try:
                            json_str = 'json://'+json.dumps(self.attrs[key], cls=JSONEncoder)
                            bb.attrs[key] = json_str
                        except:
                            raise ValueError("cannot save '%s' key in attrs dictionary" % key)

            for column, dataset in zip(columns, datasets):
                c = self[column]

                # save column attrs too
                with ff.create_from_array(dataset, c) as bb:
                    if hasattr(c, 'attrs'):
                        for key in c.attrs:
                            bb.attrs[key] = c.attrs[key]

    def read(self, columns):
        """
        Return the requested columns as dask arrays.

        Parameters
        ----------
        columns : list of str
            the names of the requested columns

        Returns
        -------
        list of :class:`dask.array.Array` :
            the list of column data, in the form of dask arrays
        """
        missing = set(columns) - set(self.columns)
        if len(missing) > 0:
            msg = "source does not contain columns: %s; " %str(missing)
            msg += "try adding columns via `source[column] = data`"
            raise ValueError(msg)

        return [self[col] for col in columns]

    def to_mesh(self, Nmesh=None, BoxSize=None, dtype='f4', interlaced=False,
                compensated=False, window='cic', weight='Weight',
                value='Value', selection='Selection', position='Position'):
        """
        Convert the CatalogSource to a MeshSource, using the specified
        parameters.

        Parameters
        ----------
        Nmesh : int, optional
            the number of cells per side on the mesh; must be provided if
            not stored in :attr:`attrs`
        BoxSize : scalar, 3-vector, optional
            the size of the box; must be provided if
            not stored in :attr:`attrs`
        dtype : string, optional
            the data type of the mesh array
        interlaced : bool, optional
            use the interlacing technique of Sefusatti et al. 2015 to reduce
            the effects of aliasing on Fourier space quantities computed
            from the mesh
        compensated : bool, optional
            whether to correct for the window introduced by the grid
            interpolation scheme
        window : str, optional
            the string specifying which window interpolation scheme to use;
            see `pmesh.window.methods`
        weight : str, optional
            the name of the column specifying the weight for each particle
        value: str, optional
            the name of the column specifying the field value for each particle
        selection : str, optional
            the name of the column that specifies which (if any) slice
            of the CatalogSource to take
        position : str, optional
            the name of the column that specifies the position data of the
            objects in the catalog

        Returns
        -------
        mesh : CatalogMesh
            a mesh object that provides an interface for gridding particle
            data onto a specified mesh
        """
        from nbodykit.base.catalogmesh import CatalogMesh
        from pmesh.window import methods

        # make sure all of the columns exist
        for col in [weight, selection]:
            if col not in self:
                raise ValueError("column '%s' missing; cannot create mesh" %col)

        if window not in methods:
            raise ValueError("valid window methods: %s" %str(methods))

        if BoxSize is None:
            try:
                BoxSize = self.attrs['BoxSize']
            except KeyError:
                raise ValueError(("cannot convert particle source to a mesh; "
                                  "'BoxSize' keyword is not supplied and the CatalogSource "
                                  "does not define one in 'attrs'."))
        if Nmesh is None:
            try:
                Nmesh = self.attrs['Nmesh']
            except KeyError:
                raise ValueError(("cannot convert particle source to a mesh; "
                                  "'Nmesh' keyword is not supplied and the CatalogSource "
                                  "does not define one in 'attrs'."))

        r = CatalogMesh(self, Nmesh=Nmesh, BoxSize=BoxSize, dtype=dtype,
                                weight=weight, selection=selection,
                                value=value, position=position)
        r.interlaced = interlaced
        r.compensated = compensated
        r.window = window
        return r


@add_metaclass(abc.ABCMeta)
class CatalogSource(CatalogSourceBase):
    """
    An abstract base class representing a catalog of discrete particles.

    This objects behaves like a structured numpy array -- it must have a
    well-defined size when initialized. The ``size`` here represents the
    number of particles in the source on the local rank.

    Subclasses of this class must define a ``size`` attribute.

    The information about each particle is stored as a series of
    columns in the format of dask arrays. These columns can be accessed
    in a dict-like fashion.

    All subclasses of this class contain the following default columns:

    #. ``Weight``
    #. ``Value``
    #. ``Selection``

    For a full description of these default columns, see
    :ref:`the documentation <catalog-source-default-columns>`.
    """
    logger = logging.getLogger('CatalogSource')

    def __init__(self, comm, use_cache=False):

        # init the base class
        CatalogSourceBase.__init__(self, comm, use_cache=use_cache)

        # if size is already computed update csize
        # otherwise the subclass shall call update_csize explicitly.
        if self.size is not NotImplemented:
            self.update_csize()

    def __repr__(self):
        size = "%d" %self.size if self.size is not NotImplemented else "NotImplemented"
        return "%s(size=%s)" %(self.__class__.__name__, size)

    def __len__(self):
        """
        The local size of the CatalogSource on a given rank.
        """
        return self.size

    def __setitem__(self, col, value):
        """
        Add columns to the CatalogSource, overriding any existing columns
        with the name ``col``.
        """
        # handle scalar values
        if numpy.isscalar(value):
            assert self.size is not NotImplemented, "size is not implemented! cannot set scalar array"
            value = ConstantArray(value, self.size, chunks=100000)

        # check the correct size, if we know the size
        if self.size is not NotImplemented:
            assert len(value) == self.size, "error setting column, data must be array of size %d" %self.size

        # call the base __setitem__
        CatalogSourceBase.__setitem__(self, col, value)

    def copy(self):
        """
        Return a copy of the CatalogSource object

        Returns
        -------
        CatalogCopy :
            the new CatalogSource object holding the copied data columns
        """
        if self.size is NotImplemented:
            return ValueError("cannot copy a CatalogSource that does not have `size` implemented")

        data = {col:self[col] for col in self.columns}
        toret = CatalogCopy(self.size, comm=self.comm, use_cache=self.use_cache, **data)
        toret.attrs.update(self.attrs)
        return toret

    @abc.abstractproperty
    def size(self):
        """
        The number of particles in the CatalogSource on the local rank.

        This property must be defined for all subclasses.
        """
        return NotImplemented

    @property
    def csize(self):
        """
        The total, collective size of the CatalogSource, i.e., summed across all
        ranks.

        It is the sum of :attr:`size` across all available ranks.
        """
        return self._csize

    @column
    def Selection(self):
        """
        A boolean column that selects a subset slice of the CatalogSource.

        By default, this column is set to ``True`` for all particles.
        """
        return ConstantArray(True, self.size, chunks=100000)

    @column
    def Weight(self):
        """
        The column giving the weight to use for each particle on the mesh.

        The mesh field is a weighted average of ``Value``, with the weights
        given by ``Weight``.

        By default, this array is set to unity for all particles.
        """
        return ConstantArray(1.0, self.size, chunks=100000)

    @column
    def Value(self):
        """
        When interpolating a CatalogSource on to a mesh, the value of this
        array is used as the Value that each particle contributes to a given
        mesh cell.

        The mesh field is a weighted average of ``Value``, with the weights
        given by ``Weight``.

        By default, this array is set to unity for all particles.
        """
        return ConstantArray(1.0, self.size, chunks=100000)

    def update_csize(self):
        """
        Set the collective size, :attr:`csize`.

        This function should be called in :func:`__init__` of a subclass,
        after :attr:`size` has been set to a valid value (not ``NotImplemented``)
        """
        if self.size is NotImplemented:
            raise ValueError(("``size`` cannot be NotImplemented when trying "
                              "to compute collective size `csize`"))

        # sum size across all ranks
        self._csize = self.comm.allreduce(self.size)

        # log some info
        if self.comm.rank == 0:
            self.logger.debug("rank 0, local number of particles = %d" %self.size)
            self.logger.info("total number of particles in %s = %d" %(str(self), self.csize))


class CatalogCopy(CatalogSource):
    """
    A CatalogSource object that holds column data copied from an
    original source

    Parameters
    ----------
    size : int
        the size of the new source; this was likely determined by
        the number of particles passing the selection criterion
    comm : MPI communicator
        the MPI communicator; this should be the same as the
        comm of the object that we are selecting from
    use_cache : bool, optional
        whether to cache results
    **columns :
        the data arrays that will be added to this source; keys
        represent the column names
    """
    def __init__(self, size, comm, use_cache=False, **columns):

        self._size = size
        CatalogSource.__init__(self, comm=comm, use_cache=use_cache)

        # store the column arrays
        for name in columns:
            self[name] = columns[name]

    @property
    def size(self):
        return self._size

def get_catalog_subset(parent, index):
    """
    Select a subset of a :class:`CatalogSource` according to a boolean
    index array.

    Returns a :class:`CatalogCopy` holding only the data that satisfies
    the slice criterion.

    Parameters
    ----------
    parent : :class:`CatalogSource`
        the parent source that will be sliced
    index : array_like
        either a dask or numpy boolean array; this determines which
        rows are included in the returned object

    Returns
    -------
    subset : :class:`CatalogCopy`
        the particle source with the same meta-data as `parent`, and
        with the sliced data arrays
    """
    # compute the index slice if needed and get the size
    if isinstance(index, da.Array):
        index = parent.compute(index)
    elif isinstance(index, list):
        index = numpy.array(index)

    # verify the index is a boolean array
    if len(index) != len(parent):
        raise ValueError("slice index has length %d; should be %d" %(len(index), len(parent)))
    if getattr(index, 'dtype', None) != '?':
        raise ValueError("index used to slice CatalogSource must be boolean and array-like")

    # new size is just number of True entries
    size = index.sum()

    # initialize subset Source of right size
    subset_data = {col:parent[col][index] for col in parent}
    toret = CatalogCopy(size, parent.comm, use_cache=parent.use_cache, **subset_data)

    # and the meta-data
    toret.attrs.update(parent.attrs)

    return toret



from .fftpower import FFTPower, ProjectedFFTPower
from .kdtree import KDDensity
from .fof import FOF
from .convpower import ConvolvedFFTPower
from .zhist import RedshiftHistogram
from .fibercollisions import FiberCollisions
from .threeptcf import Multipoles3PCF
from .paircount import SimulationBoxPairCount, SurveyDataPairCount
from .cgm import CylindricalGroups



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_allclose, assert_array_equal
import os

setup_logging("debug")
data_dir = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'data')

@MPITest([4])
def test_threeptcf(comm):

    import tempfile

    CurrentMPIComm.set(comm)
    BoxSize = 400.0

    # load the test data
    filename = os.path.join(data_dir, 'threeptcf_data.dat')
    cat = CSVCatalog(filename, names=['x', 'y', 'z', 'w'])
    cat['Position'] = transform.StackColumns(cat['x'], cat['y'], cat['z'])
    cat['Position'] *= BoxSize

    # r binning
    nbins = 8
    edges = numpy.linspace(0, 200.0, nbins+1)

    # run the algorithm
    ells = list(range(0, 11))
    r = Multipoles3PCF(cat, ells, edges, BoxSize=BoxSize, weight='w')

    # load the result from file
    truth = numpy.empty((8,8,11))
    with open(os.path.join(data_dir, 'threeptcf_result.dat'), 'r') as ff:
        for line in ff:
            fields = line.split()
            i, j = int(fields[0]), int(fields[1])
            truth[i,j,:] = list(map(float, fields[2:]))
            truth[j,i,:] = truth[i,j,:]

    # test equality
    for i, ell in enumerate(ells):
        x = r.poles['zeta_%d' %ell]
        assert_allclose(x * (4*numpy.pi)**2 / (2*ell+1), truth[...,i], rtol=1e-3, err_msg='mismatch for ell=%d' %ell)

    # save to temp file
    tmpfile = tempfile.mktemp()
    r.save(tmpfile)

    r2 = Multipoles3PCF.load(tmpfile)
    assert_array_equal(r.poles.data, r2.poles.data)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose

from nbodykit.algorithms.kdtree import KDDensity

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_kddensity(comm):
    cosmo = cosmology.Planck15

    CurrentMPIComm.set(comm)

    source = LogNormalCatalog(Plin=cosmology.LinearPower(cosmo, 0.55),
                nbar=3e-3, BoxSize=512., Nmesh=128, seed=42)

    kdden = KDDensity(source)
    assert(kdden.density.size , source.size)
    print(kdden.density.max())



import numpy
import logging
from six import string_types

from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic
from nbodykit.utils import split_size_3d
from pmesh.domain import GridND

try:
    import Corrfunc
    from Corrfunc.mocks import DDsmu_mocks
except:
    Corrfunc = None

class PairCountBase(object):
    """
    Base class for pair counting algorithms, either for a simulation box
    or survey data

    Parameters
    ----------
    mode : {'1d', '2d'}
        compute paircounts as a function of ``r`` and ``mu`` or just ``r``
    source1 : CatalogSource
        the first source of particles
    source2 : CatalogSource, optional
        the second source of particles to cross-correlate
    redges : array_like
        the radius bin edges; length of nbins+1
    Nmu : int
        the number of ``mu`` bins to use; bins range from [0,1]
    """
    def __init__(self, mode, source1, source2, redges, Nmu):

        assert mode in ['1d', '2d'], "PairCount mode must be '1d' or '2d'"
        self.source1 = source1
        self.source2 = source2
        self.comm = source1.comm

        # save the meta-data
        self.attrs = {}
        self.attrs['mode'] = mode
        self.attrs['redges'] = redges
        self.attrs['Nmu'] = Nmu

    def __getstate__(self):
        return {'result':self.result.data, 'attrs':self.attrs}

    def __setstate__(self, state):
        self.__dict__.update(state)
        redges = self.attrs['redges']
        if self.attrs['mode'] == '1d':
            self.result = BinnedStatistic(['r'], [redges], self.result, fields_to_sum=['npairs'])
        else:
            muedges = numpy.linspace(0, 1., self.attrs['Nmu']+1)
            self.result = BinnedStatistic(['r', 'mu'], [redges, muedges], self.result, fields_to_sum=['npairs'])

    def save(self, output):
        """
        Save result as a JSON file with name ``output``
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('measurement done; saving result to %s' % output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, output, comm=None):
        """
        Load a result has been saved to disk with :func:`save`.
        """
        import json
        from nbodykit.utils import JSONDecoder
        if comm.rank == 0:
            with open(output, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)
        self.__setstate__(state)
        self.comm = comm
        return self


class SimulationBoxPairCount(PairCountBase):
    """
    Count (weighted) pairs of objects in a simulation box using the
    :mod:`Corrfunc` package.

    This uses the :func:`Corrfunc.theory.DD.DD` and
    :func:`Corrfunc.theory.DDsmu.DDsmu` functions to count pairs.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~SimulationBoxPairCount.run` for the attributes storing the
    results.

    .. note::

        The algorithm expects the positions of particles in a simulation box to
        be the Cartesian ``x``, ``y``, and ``z`` vectors. To compute
        pair counts on survey data, using right ascension, declination, and
        redshift, see :class:`SurveyDataPairCount`.

    Parameters
    ----------
    mode : {'1d', '2d'}
        compute pair counts as a function of ``r`` and ``mu`` or just ``r``
    source1 : CatalogSource
        the first source of particles, providing the 'Position' column
    redges : array_like
        the radius bin edges; length of nbins+1
    BoxSize : float, 3-vector, optional
        the size of the box; if 'BoxSize' is not provided in the source
        'attrs', it must be provided here
    source2 : CatalogSource, optional
        the second source of particles to cross-correlate
    Nmu : int, optional
        the number of ``mu`` bins, ranging from 0 to 1
    los : {'x', 'y', 'z'}, int, optional
        the axis of the simulation box to treat as the line-of-sight direction;
        this can be provided as string identifying one of 'x', 'y', 'z' or
        the equivalent integer number of the axis
    periodic : bool, optional
        whether to use periodic boundary conditions
    weight : str, optional
        the name of the column in the source specifying the particle weights
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function
    """
    logger = logging.getLogger('SimulationBoxPairCount')

    def __init__(self, mode, source1, redges, BoxSize=None, periodic=True,
                    source2=None, Nmu=5, los='z', weight='Weight', **config):

        if isinstance(los, string_types):
            assert los in 'xyz', "``los`` should be one of 'x', 'y', 'z'"
            los = 'xyz'.index(los)
        elif isinstance(los, int):
            if los < 0: los += 3
            assert los in [0,1,2], "``los`` should be one of 0, 1, 2"
        else:
            raise ValueError("``los`` should be either ['x', 'y', 'z'] or [0,1,2]")

        # verify the input sources
        BoxSize = verify_input_sources(source1, source2, BoxSize, ['Position', weight])

        # init the base class
        PairCountBase.__init__(self, mode, source1, source2, redges, Nmu)

        # save the meta-data
        self.attrs['BoxSize'] = BoxSize
        self.attrs['periodic'] = periodic
        self.attrs['weight'] = weight
        self.attrs['config'] = config
        self.attrs['los'] = los

        # test Rmax for PBC
        if periodic and numpy.amax(redges) > 0.5*self.attrs['BoxSize'].min():
            raise ValueError("periodic pair counts cannot be computed for Rmax > 0.5 * BoxSize")

        # run the algorithm
        self.run()

    def run(self):
        """
        Calculate the 3D pair-counts in a simulation box as a function
        of separation ``r`` or separation and angle to line-of-sight
        (``r``, ``mu``). This adds the following attributes to the class:

        - :attr:`SimulationBoxPairCount.result`

        Attributes
        ----------
        result : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object holding the pair count and correlation
            function results. The coordinate grid is either ``r`` or
            ``r`` and ``mu``. It stores the following variables:

            - ``r``: the mean separation value in the bin
            - ``xi``: the mean correlation function value in the bin, computed as
              :math:`DD/RR - 1`, where :math:`RR` is the number of random pairs
              in the bin
            - ``npairs``: the number of pairs in the bin
            - ``weightavg``: the average weight value in the bin; each pair
              contributes the product of the individual weight values
        """
        if Corrfunc is None:
            raise ImportError(("please install Corrfunc using either ``conda install -c bccp corrfunc``"
                               " or from ``pip install pip install git+git://github.com/nickhand/Corrfunc``"))


        # some setup
        redges = self.attrs['redges']
        comm   = self.comm
        nbins  = len(redges)-1
        boxsize = self.attrs['BoxSize']
        L = None

        # determine axes perp and paralell to
        los = self.attrs['los']
        perp_axes = [0,1,2]
        perp_axes.remove(los)

        # cubic box required by Corrunc
        if self.attrs['periodic']:
            if numpy.all(boxsize == boxsize[0]):
                L = boxsize[0]
            else:
                raise ValueError(("``Corrfunc`` does not currently support periodic "
                                  "wrapping with non-cubic boxes"))

        # determine processor division for domain decomposition
        np = split_size_3d(comm.size)
        if self.comm.rank == 0:
            self.logger.info("using cpu grid decomposition: %s" %str(np))

        # get the (periodic-enforced) position
        pos1 = self.source1['Position']
        if self.attrs['periodic']:
            pos1 %= boxsize
        pos1, w1 = self.source1.compute(pos1, self.source1[self.attrs['weight']])
        N1 = comm.allreduce(len(pos1))

        if self.source2 is not None:
            pos2 = self.source2['Position']
            if self.attrs['periodic']:
                pos2 %= boxsize
            pos2, w2 = self.source2.compute(pos2, self.source2[self.attrs['weight']])
            N2 = comm.allreduce(len(pos2))
        else:
            pos2 = pos1
            w2 = w1
            N2 = N1

        # global min/max across all ranks
        posmin = numpy.asarray(comm.allgather(pos1.min(axis=0))).min(axis=0)
        posmax = numpy.asarray(comm.allgather(pos1.max(axis=0))).max(axis=0)

        # domain decomposition
        grid = [
            numpy.linspace(0, boxsize[0], np[0] + 1, endpoint=True),
            numpy.linspace(0, boxsize[1], np[1] + 1, endpoint=True),
            numpy.linspace(0, boxsize[2], np[2] + 1, endpoint=True),
        ]
        domain = GridND(grid, comm=comm)

        layout = domain.decompose(pos1, smoothing=0)
        pos1   = layout.exchange(pos1)
        w1     = layout.exchange(w1)

        # get the position/weight of the secondaries
        rmax = numpy.max(redges)
        if rmax > boxsize.max() * 0.25:
            pos2 = numpy.concatenate(comm.allgather(pos2), axis=0)
            w2   = numpy.concatenate(comm.allgather(w2), axis=0)
        else:
            layout  = domain.decompose(pos2, smoothing=rmax)
            pos2 = layout.exchange(pos2)
            w2   = layout.exchange(w2)

        # log the sizes of the trees
        self.logger.info('rank %d correlating %d x %d' %(comm.rank, len(pos1), len(pos2)))
        if comm.rank == 0: self.logger.info('all correlating %d x %d' %(N1, N2))

        # do the pair counting
        kws = {}
        kws['weights1'] = w1.astype(pos1.dtype)
        kws['periodic'] = self.attrs['periodic']
        kws['X1'] = pos1[:,perp_axes[0]]
        kws['Y1'] = pos1[:,perp_axes[1]]
        kws['Z1'] = pos1[:,los] # LOS defined with respect to this axis
        kws['X2'] = pos2[:,perp_axes[0]]
        kws['Y2'] = pos2[:,perp_axes[1]]
        kws['Z2'] = pos2[:,los]
        kws['weights2'] = w2.astype(pos2.dtype)
        kws['weight_type'] = 'pair_product'
        if L is not None: kws['boxsize'] = L

        # 1D or 2D calculations
        if self.attrs['mode'] == '1d':
            kws['output_ravg'] = True
            kws.update(self.attrs['config'])
            try:
                pc = Corrfunc.theory.DD(0, 1, redges, **kws)
            except:
                raise RuntimeError("error when calling Corrfunc.theory.DD function")
            rcol = 'ravg'
        else:
            kws['output_savg'] = True
            kws.update(self.attrs['config'])
            try:
                pc = Corrfunc.theory.DDsmu(0, 1, redges, 1.0, self.attrs['Nmu'], **kws)
            except:
                raise RuntimeError("error when calling Corrfunc.theory.DDsmu function")
            pc = pc.reshape((-1, self.attrs['Nmu']))
            rcol = 'savg'

        self.logger.info('...rank %d done correlating' %(comm.rank))

        # make a new structured array
        dtype = numpy.dtype([('r', 'f8'), ('xi', 'f8'), ('npairs', 'f8'), ('weightavg', 'f8')])
        data = numpy.empty(pc.shape, dtype=dtype)

        # do the sum across ranks
        data['r'][:] = comm.allreduce(pc['npairs']*pc[rcol])
        data['weightavg'][:] = comm.allreduce(pc['npairs']*pc['weightavg'])
        data['npairs'][:] = comm.allreduce(pc['npairs'])
        idx = data['npairs'] > 0.
        data['r'][idx] /= data['npairs'][idx]
        data['weightavg'][idx] /= data['npairs'][idx]

        # compute the random pairs from the fractional volume
        RR = 1.*N1*N2 / boxsize.prod()
        if self.attrs['mode'] == '2d':
            dr3 = numpy.diff(redges**3)
            muedges = numpy.linspace(0, 1, self.attrs['Nmu']+1)
            dmu = numpy.diff(muedges)
            RR *= 2. / 3. * numpy.pi * dr3[:,None] * dmu[None,:]
        else:
            RR *= 4. / 3. * numpy.pi * numpy.diff(redges**3)

        # correlation function value
        data['xi'] = (1. * data['npairs'] / RR) - 1.0

        # make the BinnedStatistic
        if self.attrs['mode'] == '1d':
            self.result = BinnedStatistic(['r'], [redges], data, fields_to_sum=['npairs'])
        else:
            self.result = BinnedStatistic(['r','mu'], [redges,muedges], data, fields_to_sum=['npairs'])

class SurveyDataPairCount(PairCountBase):
    """
    Count (weighted) pairs of objects from a survey data catalog using the
    :mod:`Corrfunc` package.

    This uses the:func:`Corrfunc.mocks.DDsmu_mocks.DDsmu_mocks`
    function to count pairs.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~SurveyDataPairCount.run` for the attributes storing the
    results.

    .. note::

        The algorithm expects the positions of particles from a survey catalog
        be the sky coordinates, right ascension and declination, and redshift.
        To compute pair counts in a simulation box, using the Cartesian
        coordinate vectors, see :class:`SimulationBoxPairCount`.

    .. warning::
        The right ascension and declination columns should be specified
        in degrees.

    Parameters
    ----------
    mode : {'1d', '2d'}
        compute paircounts as a function of ``r`` and ``mu`` or just ``r``
    source1 : CatalogSource
        the first source of particles, providing the 'Position' column
    redges : array_like
        the radius bin edges; length of nbins+1
    cosmo : :class:`~nbodykit.cosmology.core.Cosmology`
        the cosmology instance used to convert redshift into comoving distance
    source2 : CatalogSource, optional
        the second source of particles to cross-correlate
    Nmu : int, optional
        the number of ``mu`` bins, ranging from 0 to 1
    ra : str, optional
        the name of the column in the source specifying the
        right ascension coordinates in units of degrees; default is 'RA'
    dec : str, optional
        the name of the column in the source specifying the declination
        coordinates; default is 'DEC'
    redshift : str, optional
        the name of the column in the source specifying the redshift
        coordinates; default is 'Redshift'
    weight : str, optional
        the name of the column in the source specifying the object weights
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function
    """
    logger = logging.getLogger('SurveyDataPairCount')

    def __init__(self, mode, source1, redges, cosmo, source2=None, Nmu=5,
                    ra='RA', dec='DEC', redshift='Redshift', weight='Weight', **config):

        # verify the input sources
        verify_input_sources(source1, source2, None, [ra, dec, redshift, weight], inspect_boxsize=False)

        # init the base class
        PairCountBase.__init__(self, mode, source1, source2, redges, Nmu)

        # save the meta-data
        self.attrs['cosmo'] = cosmo
        self.attrs['weight'] = weight
        self.attrs['ra'] = ra
        self.attrs['dec'] = dec
        self.attrs['redshift'] = redshift
        self.attrs['config'] = config

        # run the algorithm
        self.run()

    def run(self):
        """
        Calculate the 3D pair-counts of a survey data catalog as a function
        of separation ``r`` or separation and angle to line-of-sight
        (``r``, ``mu``). This adds the following attribute:

        - :attr:`SurveyDataPairCount.result`

        Attributes
        ----------
        result : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object holding the pair count and correlation
            function results. The coordinate grid is either ``r`` or
            ``r`` and ``mu``. It stores the following variables:

            - ``r``: the mean separation value in the bin
            - ``npairs``: the number of pairs in the bin
            - ``weightavg``: the average weight value in the bin; each pair
              contributes the product of the individual weight values
        """
        if Corrfunc is None:
            raise ImportError(("please install Corrfunc using either ``conda install -c bccp corrfunc``"
                               " or from ``pip install pip install git+git://github.com/nickhand/Corrfunc``"))
        from nbodykit.transform import StackColumns

        # some setup
        redges = self.attrs['redges']
        comm   = self.comm
        nbins  = len(redges)-1
        poscols = [self.attrs['ra'], self.attrs['dec'], self.attrs['redshift']]
        Nmu = 1 if self.attrs['mode'] == '1d' else self.attrs['Nmu']

        # determine processor division for domain decomposition
        np = split_size_3d(comm.size)
        if self.comm.rank == 0:
            self.logger.info("using cpu grid decomposition: %s" %str(np))

        # get the (periodic-enforced) position
        pos1 = StackColumns(*[self.source1[col] for col in poscols]) # this is RA, DEC, REDSHIFT
        pos1, w1 = self.source1.compute(pos1, self.source1[self.attrs['weight']])
        rdist1 = get_comoving_dist(comm, pos1, self.attrs['cosmo'])
        pos1[:,2] = rdist1
        N1 = comm.allreduce(len(pos1))

        if self.source2 is not None:
            pos2 = StackColumns(*[self.source2[col] for col in poscols])
            pos2, w2 = self.source2.compute(pos2, self.source2[self.attrs['weight']])
            rdist2 = get_comoving_dist(comm, pos2, self.attrs['cosmo'])
            pos2[:,2] = rdist2
            N2 = comm.allreduce(len(pos2))
        else:
            pos2 = pos1
            w2 = w1
            N2 = N1

        # the source2 particles
        pos2 = numpy.concatenate(comm.allgather(pos2), axis=0)
        w2   = numpy.concatenate(comm.allgather(w2), axis=0)

        # log the sizes of the trees
        self.logger.info('rank %d correlating %d x %d' %(comm.rank, len(pos1), len(pos2)))
        if comm.rank == 0: self.logger.info('all correlating %d x %d' %(N1, N2))

        if not len(pos1) or not len(pos2):
            dtype = [('savg', 'f8'), ('npairs', 'u8'), ('weightavg', 'f8')]
            pc = numpy.zeros((nbins, Nmu), dtype=dtype)
        else:
            # do the pair counting
            kws = {}
            kws['weights1'] = w1.astype(pos1.dtype)
            kws['is_comoving_dist'] = True
            kws['RA1'] = pos1[:,0] # ra
            kws['DEC1'] = pos1[:,1] # dec
            kws['CZ1'] = pos1[:,2] # comoving distance
            kws['RA2'] = pos2[:,0]
            kws['DEC2'] = pos2[:,1]
            kws['CZ2'] = pos2[:,2]
            kws['weights2'] = w2.astype(pos2.dtype)
            kws['weight_type'] = 'pair_product'
            kws['output_savg'] = True
            kws.update(self.attrs['config'])

            try:
                pc = Corrfunc.mocks.DDsmu_mocks(0, 1, 1, Nmu, 1.0, redges, **kws)
            except:
                raise RuntimeError("error when calling Corrfunc.mocks.DDsmu_mocks function")
            pc = pc.reshape((-1, Nmu))

        self.logger.info('...rank %d done correlating' %(comm.rank))

        # make a new structured array
        dtype = numpy.dtype([('r', 'f8'), ('npairs', 'f8'), ('weightavg', 'f8')])
        data = numpy.zeros(pc.shape, dtype=dtype)

        # do the sum across ranks
        data['r'][:] = comm.allreduce(pc['npairs']*pc['savg'])
        data['weightavg'][:] = comm.allreduce(pc['npairs']*pc['weightavg'])
        data['npairs'][:] = comm.allreduce(pc['npairs'])
        idx = data['npairs'] > 0.
        data['r'][idx] /= data['npairs'][idx]
        data['weightavg'][idx] /= data['npairs'][idx]

        # make the BinnedStatistic
        if self.attrs['mode'] == '1d':
            self.result = BinnedStatistic(['r'], [redges], numpy.squeeze(data), fields_to_sum=['npairs'])
        else:
            muedges = numpy.linspace(0, 1., Nmu+1)
            self.result = BinnedStatistic(['r','mu'], [redges,muedges], data, fields_to_sum=['npairs'])

def verify_input_sources(first, second, BoxSize, required_columns, inspect_boxsize=True):
    """
    Verify the input source objects have
    """
    if second is None: second = first

    # check for comm mismatch
    assert second.comm is first.comm, "communicator mismatch between input sources"

    # check required columns
    for source in [first, second]:
        for col in required_columns:
            if col not in source:
                raise ValueError("the column '%s' is missing from input source; cannot do pair count" %col)

    if inspect_boxsize:
        _BoxSize = numpy.zeros(3)
        BoxSize1 = first.attrs.get('BoxSize', None)
        if BoxSize1 is not None:
            _BoxSize[:] = BoxSize1[:]
        if BoxSize is not None:
            _BoxSize[:] = BoxSize
        if (_BoxSize==0.).all():
            raise ValueError("BoxSize must be supplied in the source ``attr`` or via the ``BoxSize`` keyword")

        BoxSize2 = second.attrs.get('BoxSize', None)
        if BoxSize1 is not None and BoxSize2 is not None:
            if not numpy.array_equal(first.attrs['BoxSize'], BoxSize2):
                raise ValueError("BoxSize mismatch between pair count cross-correlation sources")
            if not numpy.array_equal(first.attrs['BoxSize'], _BoxSize):
                raise ValueError("BoxSize mismatch between sources and the pair count algorithm")

        return _BoxSize

def get_comoving_dist(comm, pos, cosmo):
    """
    Return comoving distances from RA, DEC, Redshift

    ``pos`` has 3 columns giving: ra, dec, redshift
    """
    ra, dec, redshift = numpy.deg2rad(pos[:,0]), numpy.deg2rad(pos[:,1]), pos[:,2]

    # compute comoving distance
    rdist = cosmo.comoving_distance(redshift) # in Mpc/h

    return rdist
